/*
   G. Wiesenekker. ZZZZZZ a chess program.
   Copyright (C) 1993-1994 G. Wiesenekker
   E-mail: wiesenekker@sara.nl

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#ifndef _ZZZZZZ_H
#define _ZZZZZZ_H

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <limits.h>
#include <time.h>
#include <math.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>
#include <sys/timeb.h>
#include <assert.h>
#ifdef _WIN32
#include <windows.h>
#endif

#ifdef _MSC_VER
#define inline __inline
#define CDECL __cdecl
#else
#define CDECL
#endif

#define VERSION     "3.51 JA"
#define BUILDDATE   "09-04-09"


static void     CDECL
                catch_sigint(int s)
{
    signal(s, catch_sigint);
}


#define     false            0
#define     FALSE            false
#define     TRUE             true
#define     true             1
#define     WHITE            1
#define     BLACK            0
#define     NO_COLOUR        2
#define     bool       	     int

static bool force_mode     = false;
static bool white_on_move  = false;
static bool black_on_move  = false;
static bool analyze_mode   = true;
static bool analyze_on     = false;  
static bool started        = false;  
static bool opponent       = false;
static bool using_setboard = false;
static int             nwhite,
                       nblack;
static int no_of_moves = 0; 
static int movecount = 0;

static char fen[265*20] = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1";
static char     *start_fen = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1";


/*
   The value of the pieces, which are the normal chess values
   1, 3, 3, 5, 10 scaled by a factor of 300 (to allow for some
   flexibility in the evaluation function parameters).
   The knight and bishop score are 3.5 pawn so that the program
   does not exchange too easily a knight or bishop for three pawns.

 */



#define PAWN_SCORE   (3 * 100)
#define KNIGHT_SCORE (3 * 350)
#define BISHOP_SCORE (3 * 350)
#define ROOK_SCORE   (3 * 520)
#define QUEEN_SCORE  (3 * 980)
#define KING_SCORE   (3 * 2000)


// material difference bonuses - JA
#define  PAWN_SCORE_BONUS      (5)  
#define  KNIGHT_SCORE_BONUS    (20)   
#define  BISHOP_SCORE_BONUS    (30)     
#define  ROOK_SCORE_BONUS      (50)    
#define  QUEEN_SCORE_BONUS     (80)    



#define NDIM 23


#ifndef DEBUG_LEVEL
#define DEBUG_LEVEL 0
#endif

#define BOARD_MAX 144
#define GET_NOTA(X) ((((X) < 0) or ((X) > BOARD_MAX)) ? "??" : nota[X])
#define GET_ROW(X) nota_row[row[X]]
#define GET_COL(X) nota_column[column[X]]

#define NBITS 8
#define NBITS_SIZE BIT(NBITS)

/*
   Some more evaluation function macro's. SCORE_LIMIT stands for infinity.
   If an endgame is detected which is won, SCORE_ENDGAME is added to the score
   to make sure the program makes a transition to the endgame.
   Of course, in some positions this leads to strange play: the program
   sacrifices a pawn instead of promoting it to a queen in order to make the
   transition to the endgame.
   If the score is greater than SCORE_ABORT, we may abort the search because
   the program found mate.
 */

#define SCORE_LIMIT   30000
#define SCORE_MATE    28000
#define SCORE_ABORT   26000
#define SCORE_ENDGAME 20000

#define MOVE_MAX 128
#define CAPT_MAX 1024

/*
   VAR_MAX is the maximum length of the principle variation.
 */

#define VAR_MAX 256
#define NODE_MAX 256

/*
   NULL_MOVE_DEPTH is the minimum search depth at which we may
   do a null move. We search regular moves at (depth - 1),
   but null moves are searched at (depth - 1 - NULL_MOVE_REDUC).
   If the null move score fails high, the node is pruned.
   (This fails terribly in case of zugzwang).
 */

#define NULL_MOVE_DEPTH  (3)      
#define NULL_MOVE_REDUC  (2)      
#define QUIES_DEPTH      (2)      

#define MARGIN (PAWN_SCORE / 2)


/*
   The program uses two hash tables: one for the alpha-beta, and one for
   the static evaluation.
   The hash table sizes are powers of two, so we can calculate the index
   by (hask_key.low & (~(~0L << CACHE_BITS))).
   Since we then already know the value of the lower CACHE_BITS bits,
   we can use these bits to store other useful information.
 */

#ifdef SMALL
#define CACHE_BITS        9
#define SCORE_CACHE_BITS  9
#else

#define CACHE_BITS        19  
#define SCORE_CACHE_BITS  20  
#endif

#define CACHE_MASK        (~(~0L << CACHE_BITS))
#define SCORE_CACHE_MASK  (~(~0L << SCORE_CACHE_BITS))

/*
   We use CACHE_DEPTH_BITS bits for the depth, so the maximum search depth than can
   be stored in the hash table is CACHE_DEPTH_MASK.
 */

#ifdef SMALL
#define CACHE_DEPTH_BITS 4
#else
#define CACHE_DEPTH_BITS 14 
#endif
#define CACHE_DEPTH_MASK (~(~0UL << CACHE_DEPTH_BITS))

/*
   CACHE_INVAL is set if the hash table entry is empty.
 */

#define CACHE_INVAL BITL(CACHE_DEPTH_BITS)
#define CACHE_LOWER BITL(CACHE_DEPTH_BITS + 1)
#define CACHE_UPPER BITL(CACHE_DEPTH_BITS + 2)
#define CACHE_true  BITL(CACHE_DEPTH_BITS + 3)

#define CACHE_KIND_BITS (CACHE_DEPTH_BITS + 4)
#define CACHE_KIND_MASK (~(~0L << CACHE_KIND_BITS))

#if CACHE_BITS < CACHE_KIND_BITS
#error CACHE_BITS should not be smaller than CACHE_KIND_BITS
#endif



/*
   We pack the move (from, to and kind) in a long.
   ADD_MOVE pushes a move on a stack. P points to the top of the stack.
 */

#define PACK_MOVE(X, K, A, B)\
        (X) = ((K) | ((B) << 8) | (A));
#define UNPACK_MOVE(X, K, A, B)\
        (K) = ((X) & 0xFFFF0000L);\
        (A) = (int) ((X) & 0x000000FFL);\
        (B) = (int) (((X) >> 8) & 0x000000FFL);
#define ADD_MOVE(P, K, A, B)\
        {\
                PACK_MOVE(*(P), K, A, B)\
                (P)++;\
        }
        
#define MAXTIME           (80)
#define RESIGN_THRESHOLD (2000)


static int  max_time         =   0;
static int  max_moves        =   0;
static int  max_depth        =   0;
static int  inc              =   0;
static int  stop_time        =   0;
static int  Best_Score       =   0;
static int  resigncounter    =   0;
static int  mynodes          =   0;

/*
   this saves some memory (the __FILE__ string is only generated once).
 */
static char *FILE_NAME       = __FILE__;



/*
   The evaluation function parameters. They just run from 'bad'
   to 'good' and increase by 5 or 10 points. Nevertheless
   the program scored 5 out of 11 in the Dutch Computerchess Championships 1993.
 */

static int       weak_queen_score = -270;
static int       weak_rook_score = -220;
static int       weak_bishop_score = -70;
static int       weak_knight_score = -120;
static int       isolated_pawn_score = -70;
static int       king_danger_score = -40;
static int       double_pawn_score = -30;
static int       lagged_pawn_score = -20;
static int       bishop_blocks_pawn_score = -15;
static int       bishop_not_developed_score = -5;
static int       knight_not_developed_score = -5;
static int       weak_pawn_attack_score = 10;
static int       pin_score = 20;
static int       rook_on_half_open_file_score = 25;
static int       knight_strong_score = 30;
static int       bishop_strong_score = 35;
static int       rook_on_open_file_score = 40;
static int       rook_on_7th_rank_score = 200;  // was 50 
static int       long_castle_score = 60;
static int       short_castle_score = 100;
static int       knight_can_move_score = 2;
static int       bishop_can_move_score = 3;
static int       rook_can_move_score = 1;


/*
   DepthMax is the maximum search depth.
   TimeLimit is the allocated time.
   GameTimeUsed is the total time used for the game.
   If UsePos is true we use the static evaluation.
 */


static int  DepthMax      = 32;
static int  GameTimeUsed  =  0;

static int  MateDepth     =  32;      
static int  UsePos        =  1;
static int  ZapCache      =  0;

/*
   dummy is always false, but the compiler does not know that, so DUMMY
   can be used to disable some compiler warnings.
 */
static int dummy          =  false;

#define VERSION_STRING_MAX 32

/*
   Setting DEBUG_LEVEL to 1 includes debugging code in the program.
   Makes it run slow, but it is very useful if you run into problems.
 */

#undef DEBUG_LEVEL

/*
   Defining SMALL reduces the sizes of some tables used by zzzzzz.
   Useful when debugging under DOS.
 */

#undef SMALL

/*
   Some general macro's used by every C programmer.
 */

#define false 0
#define true 1
#define INVALID (-1)

/*
   Cancel rediculous predefined values of LINE_MAX like 65535
 */

#undef LINE_MAX
#define LINE_MAX 256

/*
   I have had some very nasty bugs in the past by typing & instead of &&.
   
 */

#define and &&
#define or  ||


#define ABS(A) ((A) > 0 ? (A) : -(A))
#define MAX(A, B) ((A) > (B) ? (A) : (B))
#define BIT(X)  (1 << (X))
#define BITL(X) (1L << (X))
#define BITUL(X) (1UL << (X))

/*
   Leave out the error string in DOS to save precious memory.
 */


#define FATAL_ERROR(X)\
                {my_printf("??%s::%lu\n", FILE_NAME, (ulong_t) __LINE__); exit(EXIT_FAILURE);}



/*
   You will find many BUG macro's in the code. Of course, the error messages
   are only understandable if you look at the code.
 */

#define BUG(X) {if (X) FATAL_ERROR(#X)}


#define MALLOC(P, T, N)\
        {BUG(((P) = (T *) malloc(sizeof(T) * (N))) == NULL);}
#define FREE(P) {free(P);}

#define FCLOSE(F) BUG(fclose(F) == EOF)
#define FREAD(F, P, T, N) BUG(fread(P, sizeof(T), N, F) != N)
#define FREAD_STRING(F, P) {long N; FREAD(F, &N, long, 1); FREAD(F, P, char, N); }
#define FWRITE(F, P, T, N) BUG(fwrite(P, sizeof(T), N, F) != N)
#define FWRITE_STRING(F, P) {long N; N = strlen(P) + 1; FWRITE(F, &N, long, 1); FWRITE(F, P, char, N); }
#define FSEEK(F, N) BUG(fseek(F, N, SEEK_SET) != 0)
#define FBOF(F) BUG(fseek(F, 0, SEEK_SET) != 0)
#define FEOF(F) BUG(fseek(F, 0, SEEK_END) != 0)
#define FTELL(F, X) BUG(((X) = ftell(F)) < 0)

#define HUGE_MALLOC(P, T, N) BUG(((P) = (T  *) malloc(sizeof(T) * (N))) == NULL)
#define HUGE_FREE(P) {free((void  *) P);}
#define HUGE_FREAD(P, T, N, F)\
        {\
                long i;\
                T t;\
                for (i = 0; i < (N); i++)\
                {\
                        BUG(fread(&t, sizeof(T), 1, F) != 1)\
                        (P)[i] = t;\
                }\
        }
#define HUGE_FWRITE(P, T, N, F)\
        {\
                long i;\
                T t;\
                for (i = 0; i < (N); i++)\
                {\
                        t = (P)[i];\
                        BUG(fwrite(&t, sizeof(T), 1, F) != 1)\
                }\
        }


/*
   DUMMY disables some compiler warnings about unused variables or arguments
   (unused now, but may be used in the future).
   dummy is always false, but the compiler does not know that.
 */

#define DUMMY(F, V) if (dummy) my_printf(F, V);
extern int      dummy;

typedef signed char byte_t;

typedef unsigned char uchar_t;
typedef unsigned int uint_t;
typedef unsigned long ulong_t;


/*
   WHITE_COLOUR and BLACK_COLOUR to avoid name clashes with WHITE and BLACK under DOS.
 */

enum {
    WHITE_COLOUR = 0, BLACK_COLOUR, NONE, SIDE
};
static  char    *colour_as_string[2];


/*
   The following macro's define bits for the different kind of moves
   in chess.
 */

#define NULL_MOVE   BITL(16)
#define PROM_CAPT   BITL(17)
#define PROM_NORM   BITL(18)
#define CAPT        BITL(19)
#define EP          BITL(20)
#define SHORT       BITL(21)
#define LONG        BITL(22)
#define DOUBLE      BITL(23)
#define NORM        BITL(24)
#define PROM_KNIGHT BITL(25)
#define PROM_BISHOP BITL(26)
#define PROM_ROOK   BITL(27)
#define PROM_QUEEN  BITL(28)
#define INVAL_MOVE  BITL(29)

typedef int     score_t;

/*
   In a future version I will squeeze the move in an int (6 bits from,
   6 bits to, 4 bits kind).
 */

typedef long    move_t;

/*
   We use 64 bit (or 128 in case of problems) hash keys.
   See T.A. Marsland, ICCA vol. 9, no. 1 (1986) pp. 3-19.
   T. Warnock and B. Wendroff, ICCA vol. 11, no 1 (1988), pp. 10-12
 */

#undef HASH_KEY128

typedef struct {
    ulong_t         low,
                    high;
#ifdef HASH_KEY128
    ulong_t         lower,
                    higher;
#endif
}
                hash_key_t;

#ifdef HASH_KEY128
#define CLEAR_KEY(X)\
                {(X).low = (X).high = (X).lower = (X).higher = 0UL;}
#define UPDATE_KEY(X, Y)\
                {(X).low ^= (Y).low; (X).high ^= (Y).high; (X).lower ^= (Y).lower; (X).higher ^= (Y).higher;}
#define EQUAL_KEY(X, Y)\
                (((X).low == (Y).low) and ((X).high == (Y).high) and ((X).lower == (Y).lower) and ((X).higher == (Y).higher))
#define SET_KEY(X)\
                {(X).low = ulrand(); (X).high = ulrand(); (X).lower = ulrand(); (X).higher = ulrand();}
#else
#define CLEAR_KEY(X)\
                {(X).low = (X).high = 0UL;}
#define UPDATE_KEY(X, Y)\
                {(X).low ^= (Y).low; (X).high ^= (Y).high;}
#define EQUAL_KEY(X, Y)\
                (((X).low == (Y).low) and ((X).high == (Y).high))
#define SET_KEY(X)\
                {(X).low = ulrand(); (X).high = ulrand();}
#endif

/*
   We store the pieces in lists, one list for every piece, so that
   we can first generate the pawn moves, then the knight moves etc.
   (Note: I usually prefix the names of the fields in a struct with
   the name of the struct).
   board_colour is WHITE_COLOUR, BLACK_COLOUR, SIDE or NONE.
   board_kind is INVALID, PAWN, KNIGHT, ROOK, QUEEN or KING.
   If board_kind is INVALID, the end of the list is reached.
   board_value is the material value of the piece.
   board_where tells where the piece is on the board.
   if board_where is INVALID, the piece has been captured.
   So you will see often this loop:

   board_t *p;
   (loop until end of list)
   for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
   {
   (ignore captured pieces)
   if (p->board_where == INVALID) continue;
   ..
   }
   board_moved tells whether the piece has moved (useful for
   checking castling rights and pieces that have not been
   developed).
   board_captured tells whether the piece has captured (useful
   in the quiescence search).
   board_key is a pointer to a table with hash keys.
   board_dir is a pointer to a table of offsets for the
   different directions in which a piece can move.
   board_shift tells whether we may offset the piece more than
   once (false for PAWN, KNIGHT and KING, true for BISHOP,
   ROOK and QUEEN).
   board_count is a pointer to a table with piece counts.
 */

#define LIST_MAX 11

typedef struct {
    int             board_colour,
                    board_kind,
                    board_value,
                    board_where;
    int             board_moved,
                    board_captured;

    hash_key_t  *board_key;

    int             board_shift;
    int            *board_dir;
    int            *board_count;
}
                board_t;

/*
   my_colour is the colour to move. my_colour and your_colour are toggled by
   do_move and undo_move.
 */
#define WHITE_TO_MOVE {my_colour = WHITE_COLOUR; your_colour = BLACK_COLOUR;}
#define BLACK_TO_MOVE {my_colour = BLACK_COLOUR; your_colour = WHITE_COLOUR;}

static int      my_colour,
                your_colour;

/*
   A node contains all kinds of information about a node.
   We need capt and prom for undoing captures and promotions.
   We need node_moved and node_captured for resetting the
   moved and captured attributes when undoing a move.
   node_mat contains the current material balance for WHITE and BLACK.
   node_key contains the current hash key.
   node_total_key contains the total hash key (including ep or
   castling rights).
   node_reverse tells whether the move was reversible (useful for
   checking for draw by repetition).
   node_hack_ep contains the square where we may take en passant (will
   be INVALID most of the time).
   node_move is the move itself.
 */

typedef struct {
    board_t        *capt,
                   *prom;
    int             node_moved,
                    node_captured;
    hash_key_t      node_key,
                    node_total_key;
    int             node_mat[2];
    int             node_reverse;
    int             node_hack_ep;
    move_t          node_move;
}
                node_t;

/*
   tot_count contains the total piece counts for white and black.
 */

static int      tot_count[2];

extern node_t  *node;

#define SCORE_INVAL   32000

/*
   ZZZZZZ knows different move formats. long algebraic notation (e.g. g1f3)
   or standard algebraic notation (e.g. Nf3).
 */

#define MOVE_STRING_LONG 0
#define MOVE_STRING_SAN  1
#define MOVE_STRING_ANY  2

#define MOVE_STRING_MAX 16

/*
   The legal moves in the starting position are stored in array root.
   root_move is the move.
   root_score is the move score (not very useful, because it is a lower
   and upper bounds for most moves).
 */

typedef struct {
    move_t          root_move;
    score_t         root_score;
}
                root_t;

#define NROOT_MAX 256

/*
   var_t is used for printing the best variation.
 */

typedef move_t  var_t;

/*
   capt_top is a pointer to a stack where we store the moves.
 */

extern move_t  *capt_top;

extern char    *start_fen;

extern int      nroot;
extern root_t  *root;
extern int      jroot;
extern int      GameTimeUsed;


static int             ngame_moves;
static int             other_kind,other_from,other_to;


/*
   Prototypes
 */

char           *time_stamp(void);
void            my_printf(char *,...);
char           *my_fgets(char *, int, FILE *);
ulong_t         ulrand(void);
ulong_t         crc(uchar_t *);
void            handle_signal(int);
extern char     move_string[MOVE_STRING_MAX];
void            move_to_string(move_t);
void            board_to_fen(char *, int);
void            fen_to_board(char *, int);
static void            write_board(FILE * f);
void            display_score(void);
void            display_occup(void);
void            display_attack(void);
void            compact_lists(void);
int             i_am_check(int, int);
int             gen_all_moves(void);
void            do_move(move_t);
void            undo_move(void);
void            set_score(double *);
void            move_to_san_string(move_t, int);
int             search_san(char *);
int             draw_by_repetition(int);
void            gen_root(void);
static int             search_root(char *, int, int, int);
static int             do_mate(void);
static void            do_root(void);
void            sort_root(void);
static void            initialize(void);
void            pgn(void);


#endif
